@@ -1,6 +1,7 @@
 import codecs
 
 from . import v
+from vulture.utils import ExitCode
 
 assert v  # Silence pyflakes.
 
@@ -12,7 +13,7 @@ def test_encoding1(v):
 pass
 """
     )
-    assert not v.found_dead_code_or_error
+    assert v.exit_code == ExitCode.NoDeadCode
 
 
 def test_encoding2(v):
@@ -23,25 +24,24 @@ def test_encoding2(v):
 pass
 """
     )
-    assert not v.found_dead_code_or_error
+    assert v.exit_code == ExitCode.NoDeadCode
 
 
-def test_non_utf8_encoding(v, tmpdir):
+def test_non_utf8_encoding(v, tmp_path):
     code = ""
     name = "non_utf8"
-    non_utf_8_file = str(tmpdir.mkdir(name).join(name + ".py"))
+    non_utf_8_file = tmp_path / (name + ".py")
     with open(non_utf_8_file, mode="wb") as f:
         f.write(codecs.BOM_UTF16_LE)
         f.write(code.encode("utf_16_le"))
-    v.scavenge([f.name])
-    assert v.found_dead_code_or_error
+    v.scavenge([non_utf_8_file])
+    assert v.exit_code == ExitCode.InvalidInput
 
 
-def test_utf8_with_bom(v, tmpdir):
+def test_utf8_with_bom(v, tmp_path):
     name = "utf8_bom"
-    filename = str(tmpdir.mkdir(name).join(name + ".py"))
+    filepath = tmp_path / (name + ".py")
     # utf8_sig prepends the BOM to the file.
-    with open(filename, mode="w", encoding="utf-8-sig") as f:
-        f.write("")
-    v.scavenge([f.name])
-    assert not v.found_dead_code_or_error
+    filepath.write_text("", encoding="utf-8-sig")
+    v.scavenge([filepath])
+    assert v.exit_code == ExitCode.NoDeadCode
