@@ -1,7 +1,6 @@
-import argparse
 import ast
 from fnmatch import fnmatch, fnmatchcase
-import os.path
+from pathlib import Path
 import pkgutil
 import re
 import string
@@ -10,12 +9,25 @@ import sys
 from vulture import lines
 from vulture import noqa
 from vulture import utils
+from vulture.config import InputError, make_config
+from vulture.utils import ExitCode
 
-__version__ = "2.0"
 
 DEFAULT_CONFIDENCE = 60
 
 IGNORED_VARIABLE_NAMES = {"object", "self"}
+PYTEST_FUNCTION_NAMES = {
+    "setup_module",
+    "teardown_module",
+    "setup_function",
+    "teardown_function",
+}
+PYTEST_METHOD_NAMES = {
+    "setup_class",
+    "teardown_class",
+    "setup_method",
+    "teardown_method",
+}
 
 ERROR_CODES = {
     "attribute": "V101",
@@ -48,12 +60,21 @@ def _match(name, patterns, case=True):
 
 def _is_test_file(filename):
     return _match(
-        os.path.abspath(filename),
-        ["*/test/*", "*/tests/*", "*/test*.py", "*/*_test.py", "*/*-test.py"],
+        filename.resolve(),
+        ["*/test/*", "*/tests/*", "*/test*.py", "*[-_]test.py"],
         case=False,
     )
 
 
+def _assigns_special_variable__all__(node):
+    assert isinstance(node, ast.Assign)
+    return isinstance(node.value, (ast.List, ast.Tuple)) and any(
+        target.id == "__all__"
+        for target in node.targets
+        if isinstance(target, ast.Name)
+    )
+
+
 def _ignore_class(filename, class_name):
     return _is_test_file(filename) and "Test" in class_name
 
@@ -64,16 +85,20 @@ def _ignore_import(filename, import_name):
     Ignore imports from __init__.py files since they're commonly used to
     collect objects from a package.
     """
-    return os.path.basename(filename) == "__init__.py" or import_name == "*"
+    return filename.name == "__init__.py" or import_name == "*"
 
 
 def _ignore_function(filename, function_name):
-    return function_name.startswith("test_") and _is_test_file(filename)
+    return (
+        function_name in PYTEST_FUNCTION_NAMES
+        or function_name.startswith("test_")
+    ) and _is_test_file(filename)
 
 
 def _ignore_method(filename, method_name):
     return _is_special_name(method_name) or (
-        method_name.startswith("test_") and _is_test_file(filename)
+        (method_name in PYTEST_METHOD_NAMES or method_name.startswith("test_"))
+        and _is_test_file(filename)
     )
 
 
@@ -177,9 +202,6 @@ class Vulture(ast.NodeVisitor):
         def get_list(typ):
             return utils.LoggingList(typ, self.verbose)
 
-        def get_set(typ):
-            return utils.LoggingSet(typ, self.verbose)
-
         self.defined_attrs = get_list("attribute")
         self.defined_classes = get_list("class")
         self.defined_funcs = get_list("function")
@@ -189,44 +211,44 @@ class Vulture(ast.NodeVisitor):
         self.defined_vars = get_list("variable")
         self.unreachable_code = get_list("unreachable_code")
 
-        self.used_attrs = get_set("attribute")
-        self.used_names = get_set("name")
+        self.used_names = utils.LoggingSet("name", self.verbose)
 
         self.ignore_names = ignore_names or []
         self.ignore_decorators = ignore_decorators or []
 
-        self.filename = ""
+        self.filename = Path()
         self.code = []
-        self.found_dead_code_or_error = False
+        self.exit_code = ExitCode.NoDeadCode
 
     def scan(self, code, filename=""):
+        filename = Path(filename)
         self.code = code.splitlines()
         self.noqa_lines = noqa.parse_noqa(self.code)
         self.filename = filename
 
         def handle_syntax_error(e):
             text = f' at "{e.text.strip()}"' if e.text else ""
-            print(
+            self._log(
                 f"{utils.format_path(filename)}:{e.lineno}: {e.msg}{text}",
                 file=sys.stderr,
+                force=True,
             )
-            self.found_dead_code_or_error = True
+            self.exit_code = ExitCode.InvalidInput
 
         try:
-            node = (
-                ast.parse(code, filename=self.filename, type_comments=True)
-                if sys.version_info >= (3, 8)  # type_comments requires 3.8+
-                else ast.parse(code, filename=self.filename)
+            node = ast.parse(
+                code, filename=str(self.filename), type_comments=True
             )
         except SyntaxError as err:
             handle_syntax_error(err)
         except ValueError as err:
             # ValueError is raised if source contains null bytes.
-            print(
+            self._log(
                 f'{utils.format_path(filename)}: invalid source code "{err}"',
                 file=sys.stderr,
+                force=True,
             )
-            self.found_dead_code_or_error = True
+            self.exit_code = ExitCode.InvalidInput
         else:
             # When parsing type comments, visiting can throw SyntaxError.
             try:
@@ -236,17 +258,19 @@ class Vulture(ast.NodeVisitor):
 
     def scavenge(self, paths, exclude=None):
         def prepare_pattern(pattern):
-            if not any(char in pattern for char in ["*", "?", "["]):
+            if not any(char in pattern for char in "*?["):
                 pattern = f"*{pattern}*"
             return pattern
 
         exclude = [prepare_pattern(pattern) for pattern in (exclude or [])]
 
-        def exclude_file(name):
-            return any(fnmatch(name, pattern) for pattern in exclude)
+        def exclude_path(path):
+            return _match(path, exclude, case=False)
+
+        paths = [Path(path) for path in paths]
 
         for module in utils.get_modules(paths):
-            if exclude_file(module):
+            if exclude_path(module):
                 self._log("Excluded:", module)
                 continue
 
@@ -254,23 +278,24 @@ class Vulture(ast.NodeVisitor):
             try:
                 module_string = utils.read_file(module)
             except utils.VultureInputException as err:  # noqa: F841
-                print(
+                self._log(
                     f"Error: Could not read file {module} - {err}\n"
                     f"Try to change the encoding to UTF-8.",
                     file=sys.stderr,
+                    force=True,
                 )
-                self.found_dead_code_or_error = True
+                self.exit_code = ExitCode.InvalidInput
             else:
                 self.scan(module_string, filename=module)
 
         unique_imports = {item.name for item in self.defined_imports}
         for import_name in unique_imports:
-            path = os.path.join("whitelists", import_name) + "_whitelist.py"
-            if exclude_file(path):
+            path = Path("whitelists") / (import_name + "_whitelist.py")
+            if exclude_path(path):
                 self._log("Excluded whitelist:", path)
             else:
                 try:
-                    module_data = pkgutil.get_data("vulture", path)
+                    module_data = pkgutil.get_data("vulture", str(path))
                     self._log("Included whitelist:", path)
                 except OSError:
                     # Most imported modules don't have a whitelist.
@@ -286,7 +311,7 @@ class Vulture(ast.NodeVisitor):
             raise ValueError("min_confidence must be between 0 and 100.")
 
         def by_name(item):
-            return (item.filename.lower(), item.first_lineno)
+            return (str(item.filename).lower(), item.first_lineno)
 
         def by_size(item):
             return (item.size,) + by_name(item)
@@ -319,53 +344,52 @@ class Vulture(ast.NodeVisitor):
         for item in self.get_unused_code(
             min_confidence=min_confidence, sort_by_size=sort_by_size
         ):
-            print(
+            self._log(
                 item.get_whitelist_string()
                 if make_whitelist
-                else item.get_report(add_size=sort_by_size)
+                else item.get_report(add_size=sort_by_size),
+                force=True,
             )
-            self.found_dead_code_or_error = True
-        return self.found_dead_code_or_error
+            self.exit_code = ExitCode.DeadCode
+        return self.exit_code
 
     @property
     def unused_classes(self):
-        return _get_unused_items(
-            self.defined_classes, self.used_attrs | self.used_names
-        )
+        return _get_unused_items(self.defined_classes, self.used_names)
 
     @property
     def unused_funcs(self):
-        return _get_unused_items(
-            self.defined_funcs, self.used_attrs | self.used_names
-        )
+        return _get_unused_items(self.defined_funcs, self.used_names)
 
     @property
     def unused_imports(self):
-        return _get_unused_items(
-            self.defined_imports, self.used_names | self.used_attrs
-        )
+        return _get_unused_items(self.defined_imports, self.used_names)
 
     @property
     def unused_methods(self):
-        return _get_unused_items(self.defined_methods, self.used_attrs)
+        return _get_unused_items(self.defined_methods, self.used_names)
 
     @property
     def unused_props(self):
-        return _get_unused_items(self.defined_props, self.used_attrs)
+        return _get_unused_items(self.defined_props, self.used_names)
 
     @property
     def unused_vars(self):
-        return _get_unused_items(
-            self.defined_vars, self.used_attrs | self.used_names
-        )
+        return _get_unused_items(self.defined_vars, self.used_names)
 
     @property
     def unused_attrs(self):
-        return _get_unused_items(self.defined_attrs, self.used_attrs)
+        return _get_unused_items(self.defined_attrs, self.used_names)
 
-    def _log(self, *args):
-        if self.verbose:
-            print(*args)
+    def _log(self, *args, file=None, force=False):
+        if self.verbose or force:
+            file = file or sys.stdout
+            try:
+                print(*args, file=file)
+            except UnicodeEncodeError:
+                # Some terminals can't print Unicode symbols.
+                x = " ".join(map(str, args))
+                print(x.encode(), file=file)
 
     def _add_aliases(self, node):
         """
@@ -429,40 +453,6 @@ class Vulture(ast.NodeVisitor):
                     confidence=100,
                 )
 
-    def _handle_string(self, s):
-        """
-        Parse variable names in format strings:
-
-        '%(my_var)s' % locals()
-        '{my_var}'.format(**locals())
-        f'{my_var}'
-
-        """
-        # Old format strings.
-        self.used_names |= set(re.findall(r"\%\((\w+)\)", s))
-
-        def is_identifier(name):
-            return bool(re.match(r"[a-zA-Z_][a-zA-Z0-9_]*", name))
-
-        # New format strings.
-        parser = string.Formatter()
-        try:
-            names = [name for _, name, _, _ in parser.parse(s) if name]
-        except ValueError:
-            # Invalid format string.
-            names = []
-
-        for field_name in names:
-            # Remove brackets and contents: "a[0][b].c[d].e" -> "a.c.e".
-            # "a.b.c" -> name = "a", attributes = ["b", "c"]
-            name_and_attrs = re.sub(r"\[\w*\]", "", field_name).split(".")
-            name = name_and_attrs[0]
-            if is_identifier(name):
-                self.used_names.add(name)
-            for attr in name_and_attrs[1:]:
-                if is_identifier(attr):
-                    self.used_attrs.add(attr)
-
     def _define(
         self,
         collection,
@@ -487,14 +477,13 @@ class Vulture(ast.NodeVisitor):
         if ignored(first_lineno):
             self._log(f'Ignoring {typ} "{name}"')
         else:
-            last_lineno = lines.get_last_line_number(last_node)
             collection.append(
                 Item(
                     name,
                     typ,
                     self.filename,
                     first_lineno,
-                    last_lineno,
+                    lines.get_last_line_number(last_node),
                     message=message,
                     confidence=confidence,
                 )
@@ -520,7 +509,73 @@ class Vulture(ast.NodeVisitor):
         if isinstance(node.ctx, ast.Store):
             self._define(self.defined_attrs, node.attr, node)
         elif isinstance(node.ctx, ast.Load):
-            self.used_attrs.add(node.attr)
+            self.used_names.add(node.attr)
+
+    def visit_BinOp(self, node):
+        """
+        Parse variable names in old format strings:
+
+        "%(my_var)s" % locals()
+        """
+        if (
+            utils.is_ast_string(node.left)
+            and isinstance(node.op, ast.Mod)
+            and self._is_locals_call(node.right)
+        ):
+            self.used_names |= set(re.findall(r"%\((\w+)\)", node.left.value))
+
+    def visit_Call(self, node):
+        # Count getattr/hasattr(x, "some_attr", ...) as usage of some_attr.
+        if isinstance(node.func, ast.Name) and (
+            (node.func.id == "getattr" and 2 <= len(node.args) <= 3)
+            or (node.func.id == "hasattr" and len(node.args) == 2)
+        ):
+            attr_name_arg = node.args[1]
+            if utils.is_ast_string(attr_name_arg):
+                self.used_names.add(attr_name_arg.value)
+
+        # Parse variable names in new format strings:
+        # "{my_var}".format(**locals())
+        if (
+            isinstance(node.func, ast.Attribute)
+            and utils.is_ast_string(node.func.value)
+            and node.func.attr == "format"
+            and any(
+                kw.arg is None and self._is_locals_call(kw.value)
+                for kw in node.keywords
+            )
+        ):
+            self._handle_new_format_string(node.func.value.value)
+
+    def _handle_new_format_string(self, s):
+        def is_identifier(name):
+            return bool(re.match(r"[a-zA-Z_][a-zA-Z0-9_]*", name))
+
+        parser = string.Formatter()
+        try:
+            names = [name for _, name, _, _ in parser.parse(s) if name]
+        except ValueError:
+            # Invalid format string.
+            names = []
+
+        for field_name in names:
+            # Remove brackets and their contents: "a[0][b].c[d].e" -> "a.c.e",
+            # then split the resulting string: "a.b.c" -> ["a", "b", "c"]
+            vars = re.sub(r"\[\w*\]", "", field_name).split(".")
+            for var in vars:
+                if is_identifier(var):
+                    self.used_names.add(var)
+
+    @staticmethod
+    def _is_locals_call(node):
+        """Return True if the node is `locals()`."""
+        return (
+            isinstance(node, ast.Call)
+            and isinstance(node.func, ast.Name)
+            and node.func.id == "locals"
+            and not node.args
+            and not node.keywords
+        )
 
     def visit_ClassDef(self, node):
         for decorator in node.decorator_list:
@@ -585,27 +640,27 @@ class Vulture(ast.NodeVisitor):
 
     def visit_Name(self, node):
         if (
-            isinstance(node.ctx, ast.Load)
+            isinstance(node.ctx, (ast.Load, ast.Del))
             and node.id not in IGNORED_VARIABLE_NAMES
         ):
             self.used_names.add(node.id)
         elif isinstance(node.ctx, (ast.Param, ast.Store)):
             self._define_variable(node.id, node)
 
-    if sys.version_info < (3, 8):
-
-        def visit_Str(self, node):
-            self._handle_string(node.s)
-
-    else:
-
-        def visit_Constant(self, node):
-            if isinstance(node.value, str):
-                self._handle_string(node.value)
+    def visit_Assign(self, node):
+        if _assigns_special_variable__all__(node):
+            assert isinstance(node.value, (ast.List, ast.Tuple))
+            for elt in node.value.elts:
+                if utils.is_ast_string(elt):
+                    self.used_names.add(elt.value)
 
     def visit_While(self, node):
         self._handle_conditional_node(node, "while")
 
+    def visit_MatchClass(self, node):
+        for kwd_attr in node.kwd_attrs:
+            self.used_names.add(kwd_attr)
+
     def visit(self, node):
         method = "visit_" + node.__class__.__name__
         visitor = getattr(self, method, None)
@@ -666,80 +721,23 @@ class Vulture(ast.NodeVisitor):
                 self.visit(value)
 
 
-def _parse_args():
-    def csv(exclude):
-        return exclude.split(",")
-
-    usage = "%(prog)s [options] PATH [PATH ...]"
-    version = f"vulture {__version__}"
-    glob_help = "Patterns may contain glob wildcards (*, ?, [abc], [!abc])."
-    parser = argparse.ArgumentParser(prog="vulture", usage=usage)
-    parser.add_argument(
-        "paths",
-        nargs="+",
-        metavar="PATH",
-        help="Paths may be Python files or directories. For each directory"
-        " Vulture analyzes all contained *.py files.",
-    )
-    parser.add_argument(
-        "--exclude",
-        metavar="PATTERNS",
-        type=csv,
-        help=f"Comma-separated list of paths to ignore (e.g.,"
-        f' "*settings.py,docs/*.py"). {glob_help} A PATTERN without glob'
-        f" wildcards is treated as *PATTERN*.",
-    )
-    parser.add_argument(
-        "--ignore-decorators",
-        metavar="PATTERNS",
-        type=csv,
-        help=f"Comma-separated list of decorators. Functions and classes using"
-        f' these decorators are ignored (e.g., "@app.route,@require_*").'
-        f" {glob_help}",
-    )
-    parser.add_argument(
-        "--ignore-names",
-        metavar="PATTERNS",
-        type=csv,
-        default=None,
-        help=f'Comma-separated list of names to ignore (e.g., "visit_*,do_*").'
-        f" {glob_help}",
-    )
-    parser.add_argument(
-        "--make-whitelist",
-        action="store_true",
-        help="Report unused code in a format that can be added to a"
-        " whitelist module.",
-    )
-    parser.add_argument(
-        "--min-confidence",
-        type=int,
-        default=0,
-        help="Minimum confidence (between 0 and 100) for code to be"
-        " reported as unused.",
-    )
-    parser.add_argument(
-        "--sort-by-size",
-        action="store_true",
-        help="Sort unused functions and classes by their lines of code.",
-    )
-    parser.add_argument("-v", "--verbose", action="store_true")
-    parser.add_argument("--version", action="version", version=version)
-    return parser.parse_args()
-
-
 def main():
-    args = _parse_args()
+    try:
+        config = make_config()
+    except InputError as e:
+        print(e, file=sys.stderr)
+        sys.exit(ExitCode.InvalidCmdlineArguments)
+
     vulture = Vulture(
-        verbose=args.verbose,
-        ignore_names=args.ignore_names,
-        ignore_decorators=args.ignore_decorators,
+        verbose=config["verbose"],
+        ignore_names=config["ignore_names"],
+        ignore_decorators=config["ignore_decorators"],
     )
-    vulture.scavenge(args.paths, exclude=args.exclude)
+    vulture.scavenge(config["paths"], exclude=config["exclude"])
     sys.exit(
         vulture.report(
-            min_confidence=args.min_confidence,
-            sort_by_size=args.sort_by_size,
-            make_whitelist=args.make_whitelist,
+            min_confidence=config["min_confidence"],
+            sort_by_size=config["sort_by_size"],
+            make_whitelist=config["make_whitelist"],
         )
     )
